#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

#############################################################################
#
# G-code backtracer
# copyright 2006-2007 by Fuzzball Software
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
#############################################################################

#############################################################################
#
# Modified by Jaroslaw Krzywicki krzywy77@gmail.com - april/may 2007
#
#############################################################################

package require opt
#catch { package require Img }

##### New code J.K.
proc gcode_toolfile_open {base} {
    set filetypes {
        {{Text Files}	{.txt}	TEXT	}
	{{Any Files}	{*}	TEXT	}
    }
    set file [tk_getOpenFile -defaultextension .txt -filetypes $filetypes \
		  -title "Open toolfile" -parent $base]
    if {$file != ""} {
	gcode_tools_load $file
    }
}

proc gcode_print_3view {base} {
    set isocanv $base.isocanv
    set filetypes {
        {{Postscript Files}	{.ps}	TEXT	}
    }

    set file [tk_getSaveFile -defaultextension .ps -filetypes $filetypes \
		  -title "Save view 3D" -parent $base]
    if {$file != ""} {
	$isocanv postscript -file $file
    }
    #puts "Ho ho ho!"

}


##### / J.K.

proc gcode_tools_load {toolfile} {

##### Modified code J.K.
	catch {
		set f [open $toolfile "r"]

		while {![eof $f]} {
			set line [string trim [gets $f]]
			if {$line == ""} {
				continue
			}
			scan $line "%d %d %f %f %s" pocket fms tlo diam comment

			puts "$pocket, $fms, $tlo, $diam"

			gcode_tool_define $pocket $diam $tlo
		}
	}
	catch {
		close $f
	}
#### / J. K.
}


proc gcode_tool_define {toolnum tooldiam toollen} {
	global gcode_tools
	set gcode_tools($toolnum-DIAM) $tooldiam
	set gcode_tools($toolnum-LENGTH) $toollen
}


proc gcode_tool_diam {toolnum} {
	global gcode_tools
	if {![info exists gcode_tools($toolnum-DIAM)]} {
		error "Tool not defined."
	}
	return $gcode_tools($toolnum-DIAM)
}


proc gcode_tool_length {toolnum} {
	global gcode_tools
	if {![info exists gcode_tools($toolnum-LENGTH)]} {
		error "Tool not defined."
	}
	return $gcode_tools($toolnum-LENGTH)
}


proc gcode_parameter {parmnum linenum} {
	global gcode_vars
	if {$parmnum < 1 || $parmnum > 5399} {
		error "Bad parameter number after # at line $linenum"
	}
	if {abs($parmnum-int($parmnum+0.5)) > 0.0001} {
		error "Bad parameter number #$parmnum at line $linenum"
	}
	if {![info exists gcode_vars($parmnum)]} {
		if {$parmnum == 5220} {
			return 1.0
		} else {
			return 0.0
		}
	}
	return $gcode_vars($parmnum)
}


proc gcode_parameter_set {parmnum value linenum} {
	global gcode_vars_journal
	if {$parmnum < 1 || $parmnum > 5399} {
		error "Bad parameter number #$parmnum at line $linenum"
	}
	if {abs($parmnum-int($parmnum+0.5)) > 0.0001} {
		error "Bad parameter number #$parmnum at line $linenum"
	}
	if {![info exists gcode_vars_journal]} {
		set gcode_vars_journal {}
	}
	lappend gcode_vars_journal $parmnum $value
}


proc gcode_parameters_commit {} {
	global gcode_vars gcode_vars_journal
	if {![info exists gcode_vars_journal]} {
		set gcode_vars_journal {}
	}
	foreach {parmnum value} $gcode_vars_journal {
		set gcode_vars($parmnum) $value
	}
}


proc gcode_integer {strvar cmd linenum} {
	upvar $strvar str
	if {$str == ""} {
		error "Expected integer or expression after '$cmd' at line $linenum"
	}
	set val [gcode_number str $cmd $linenum]
	set intval [expr {round($val)}]
	if {abs($val-$intval) < 0.00001} {
		return $intval
	}
	error "Expected integer after '$cmd' at line $linenum"
}


proc gcode_expr {strvar cmd linenum} {
	upvar $strvar str
	set pi 3.141592653589793236
	set preconv 1.0
	set postconv 1.0
	while {[string index $str 0] == "-"} {
		set postconv [expr {-1.0*$postconv}]
		set str [string range $str 1 end]
	}
	if {[string is alpha [string index $str 0]]} {
		set word {}
		while {[string is alpha [string index $str 0]]} {
			append word [string index $str 0]
			set str [string range $str 1 end]
		}
		set word [string tolower $word]
		switch -exact -- $word {
			"abs" -
			"exp" -
			"round" -
			"sqrt" {
				set op $word
			}
			"cos" -
			"sin" -
			"tan" {
				set op $word
				set preconv [expr {$pi/180.0}]
			}
			"acos" -
			"asin" {
				set op $word
				set postconv [expr {$postconv*180.0/$pi}]
			}
			"atan" {
				set op "atan2"
				set postconv [expr {$postconv*180.0/$pi}]
			}
			"ln" {
				set op "log"
			}
			"fix" {
				set op "floor"
			}
			"fup" {
				set op "ceil"
			}
			default {
				error "Unrecognized word '$word' in line $linenum"
			}
		}
		if {[string index $str 0] != "\["} {
			error "Expected \[ after '$word' in line $linenum"
		}
		set val1 [gcode_expr str $cmd $linenum]
		if {$word == "atan"} {
			if {[string index $str 0] != "/"} {
				error "Expected '/ between arguments of '$word' in line $linenum"
			}
			set str [string range $str 1 end]
			if {[string index $str 0] != "\["} {
				error "Expected \[ after / for '$word' in line $linenum"
			}
			set val2 [gcode_expr str $cmd $linenum]
			set result [expr "${op}($val2,$val1)"]
		} else {
			set val1 [expr {$val1*$preconv}]
			set result [expr "${op}($val1)"]
		}
	} else {
		# Not a unary operator
		set nch [string index $str 0]
		if {[string is digit $nch] || $nch == "." || $nch == "#" || $nch == "\["} {
			# It's a number or parameter or sub-expression.
			set result [gcode_number str $cmd $linenum]
		} else {
			error "Unexpected operator '$ch' in expression at line $linenum"
		}
	}
	set result [expr {$result*$postconv}]

	# Check for binary operators
	set ops {}
	lappend ops [list 1 "+" 0.0]
	lappend ops [list 1 "+" $result]
	while {[string index $str 0] != "\]"} {
		set op3 [string index $str 0]
		if {[string range $str 0 1] == "**"} {
			set str [string range $str 2 "end"]
			set op3 "**"
			set pri3 3
		} elseif {[string toupper [string range $str 0 2]] == "MOD"} {
			set str [string range $str 3 "end"]
			set op3 "%"
			set pri3 2
		} elseif {$op3 == "*" || $op3 == "/"} {
			set str [string trimleft [string range $str 1 "end"]]
			set pri3 2
		} elseif {$op3 == "+" || $op3 == "-"} {
			set str [string trimleft [string range $str 1 "end"]]
			set pri3 1
		} else {
			error "Unrecognized operator '$currop' in line $linenum"
		}
		foreach {pri1 op1 val1} [lindex $ops end-1] break
		foreach {pri2 op2 val2} [lindex $ops end] break
		set val3 [gcode_expr str $cmd $linenum]
		lappend ops [list $pri3 $op3 $val3]
		while {$pri2 >= $pri3 && [llength $ops] > 2} {
			if {$op2 == "**"} {
				set val1 [expr {pow($val1,$val2)}]
			} else {
				set val1 [expr "$val1$op2$val2"]
			}
			set ops [lreplace $ops "end-2" "end-1" [list $pri1 $op1 $val1]]
			foreach {pri1 op1 val1} [lindex $ops end-2] break
			foreach {pri2 op2 val2} [lindex $ops end-1] break
		}
	}
	while {[llength $ops] > 1} {
		foreach {pri1 op1 val1} [lindex $ops end-1] break
		foreach {pri2 op2 val2} [lindex $ops end] break
		if {$op2 == "**"} {
			set result [expr {pow($val1,$val2)}]
		} else {
			set result [expr "$val1$op2$val2"]
		}
		set ops [lreplace $ops "end-1" "end" [list $pri1 $op1 $result]]
	}
	return [lindex [lindex $ops 0] 2]
}


proc gcode_number {strvar cmd linenum} {
	upvar $strvar str
	if {$str == ""} {
		error "Expected number after '$cmd' at line $linenum"
	}

##### New code J.K.
    if {$cmd =="N"} {
	set indx 0
	while { [string index $str $indx] == "0" } {
	    set indx [expr $indx + 1]
 	}
	set indxch [string index $str $indx]
	if { [string is digit $indxch] == 1 } {
	    set str [string replace $str 0 [expr $indx - 1]]
	} else {
	    set str [string replace $str 0 [expr $indx - 2]]
	}
    }
##### / J.K.

	set nch [string index $str 0]
	if {$nch == "+" || $nch == "-" || [string is digit $nch] || $nch == "."} {
		set postmult 1.0
		while {$nch == "+" || $nch == "-"} {
			if {$nch == "-"} {
				set postmult [expr {0.0-$postmult}]
			}
			set str [string trimleft [string range $str 1 end]]
			set nch [string index $str 0]
		}
		set num {}
		set havedec 0
		while {$str != "" && ([string is digit $nch] || $nch == ".")} {
			if {[string is digit $nch]} {
				append num $nch
			} elseif {$nch == "."} {
				if {$havedec} {
					error "Malformed number after '$cmd' in line $linenum"
				}
				set havedec 1
				append num $nch
			}
			set str [string trimleft [string range $str 1 end]]
			set nch [string index $str 0]
		}
		if {$num == "" || $num == "."} {
			error "Malformed number after '$cmd' in line $linenum"
		}
		set val [expr {$postmult*$num}]
		return $val
	} elseif {$nch == "\["} {
		set str [string trimleft [string range $str 1 end]]
		if {[string index $str 0] == "\]"} {
			error "Expected expression inside \[\] in line $linenum"
		}
		set result [gcode_expr str $cmd $linenum]
		if {[string index $str 0] != "\]"} {
			error "Expected \] after expression in line $linenum"
		}
		set str [string trimleft [string range $str 1 end]]
		return $result
	} elseif {$nch == "#"} {
		set str [string trimleft [string range $str 1 end]]
		set parmnum [gcode_integer str "#" $linenum]
		return [gcode_parameter $parmnum $linenum]
	}
	error "Expected number after '$cmd' at line $linenum"
}


proc matrix_mult {mat1 mat2} {
	set m [llength $mat1]
	set n [llength [lindex $mat1 0]]
	set o [llength $mat1]
	set p [llength [lindex $mat2 0]]
	if {$n != $o} {
		error "Cols in mat1 must = rows in mat2"
	}
	set mat3 {}
	for {set i 0} {$i < $m} {incr i} {
		set row {}
		for {set j 0} {$j < $p} {incr j} {
			set sum 0.0
			for {set r 0} {$r < $n} {incr r} {
				set sum [expr {$sum+[lindex [lindex $mat1 $i] $r]*[lindex [lindex $mat2 $r] $j]}]
			}
			lappend row $sum
		}
		lappend mat3 $row
	}
	return $mat3
}


proc gcode_rotation_matrix {dx dy dz ang} {
	set pi 3.141592653589793236
	set ang [expr {$ang*$pi/180.0}]
	set mat [list \
		[list [expr {(1-cos($ang))*$dx*$dx+cos($ang)}]     [expr {(1-cos($ang))*$dy*$dx-sin($ang)*$dz}] [expr {(1-cos($ang))*$dz*$dx+sin($ang)*$dy}]] \
		[list [expr {(1-cos($ang))*$dx*$dy+sin($ang)*$dz}] [expr {(1-cos($ang))*$dy*$dy+cos($ang)}]     [expr {(1-cos($ang))*$dz*$dy-sin($ang)*$dx}]] \
		[list [expr {(1-cos($ang))*$dx*$dz-sin($ang)*$dy}] [expr {(1-cos($ang))*$dy*$dz+sin($ang)*$dx}] [expr {(1-cos($ang))*$dz*$dz+cos($ang)}]    ] \
	]
	return $mat
}


proc gcode_shift_line {x1 y1 x2 y2 amount} {
    set pi 3.141592653589793236
    set cx [expr {($x1+$x2)/2.0}]
    set cy [expr {($y1+$y2)/2.0}]
    set dx [expr {$x1-$cx}]
    set dy [expr {$y1-$cy}]
    set oang [expr {atan2($dy,$dx)}]
    set pang [expr {atan2($dy,$dx)+($pi/2.0)}]
    set rad [expr {sqrt($dx*$dx+$dy*$dy)}]
    set ex [expr {$cx+$amount*cos($pang)}]
    set ey [expr {$cy+$amount*sin($pang)}]
    set px [expr {$ex+$rad*cos($oang)}]
    set py [expr {$ey+$rad*sin($oang)}]
    return [list $px $py]
}


proc gcode_toolpath_append {var x y z a b c type linenum} {
	global xoffset yoffset zoffset
	global aoffset boffset coffset
	global prev_x prev_y prev_z
	global prev_a prev_b prev_c
	global a_matrix b_matrix c_matrix
	global vector_ax0 vector_ay0 vector_az0
	global vector_ax1 vector_ay1 vector_az1
	global vector_bx0 vector_by0 vector_bz0
	global vector_bx1 vector_by1 vector_bz1
	global vector_cx0 vector_cy0 vector_cz0
	global vector_cx1 vector_cy1 vector_cz1
	global lencomp lencomptool
	global diamcomp diamcomptool
	global tool_width

	upvar $var coordlist

	set lastx $x
	set lasty $y
	set lastz $z
	set lasta $a
	set lastb $b
	set lastc $c
	set coordsysnum [gcode_parameter 5220 $linenum]
	if {$coordsysnum > 0} {
		set parmbase [expr {5221+20*($coordsysnum-1)}]
		set lastx [expr {$lastx+[gcode_parameter [expr {$parmbase+0}] $linenum]}]
		set lasty [expr {$lasty+[gcode_parameter [expr {$parmbase+1}] $linenum]}]
		set lastz [expr {$lastz+[gcode_parameter [expr {$parmbase+2}] $linenum]}]
		set lasta [expr {$lasta+[gcode_parameter [expr {$parmbase+3}] $linenum]}]
		set lastb [expr {$lastb+[gcode_parameter [expr {$parmbase+4}] $linenum]}]
		set lastc [expr {$lastc+[gcode_parameter [expr {$parmbase+5}] $linenum]}]
	}
	set lastx [expr {$lastx+$xoffset}]
	set lasty [expr {$lasty+$yoffset}]
	set lastz [expr {$lastz+$zoffset}]
	set lasta [expr {$lasta+$aoffset}]
	set lastb [expr {$lastb+$boffset}]
	set lastc [expr {$lastc+$coffset}]

	set lastx2 $lastx
	set lasty2 $lasty

	if {$lencomp == "plus"} {
		set tlo [gcode_tool_length $lencomptool]
		set lastz [expr {$lastz-$tlo}]
	}

	if {$diamcomp == "left"} {
		set diam [gcode_tool_diam $diamcomptool]
		foreach {lastx2 lasty2} [gcode_shift_line $prev_x $prev_y $lastx2 $lasty2 [expr {$diam/-2.0}]] break
	} elseif {$diamcomp == "right"} {
		set diam [gcode_tool_diam $diamcomptool]
		foreach {lastx2 lasty2} [gcode_shift_line $prev_x $prev_y $lastx2 $lasty2 [expr {$diam/2.0}]] break
	}

	set steps 1
	if {abs($lasta-$prev_a) > 5.0} {
		set est_steps [expr {int(abs(($lasta-$prev_a)/5.0))}]
		if {$est_steps > $steps} {
			set steps $est_steps
		}
	}
	if {abs($lastb-$prev_b) > 5.0} {
		set est_steps [expr {int(abs(($lastb-$prev_b)/5.0))}]
		if {$est_steps > $steps} {
			set steps $est_steps
		}
	}
	if {abs($lastc-$prev_c) > 5.0} {
		set est_steps [expr {int(abs(($lastc-$prev_c)/5.0))}]
		if {$est_steps > $steps} {
			set steps $est_steps
		}
	}
	set partval [expr {1.0/(0.0+$steps)}]

	for {set i 1} {$i <= $steps} {incr i} {
		set realx [expr {$prev_x+($lastx2-$prev_x)*$partval*$i}]
		set realy [expr {$prev_y+($lasty2-$prev_y)*$partval*$i}]
		set realz [expr {$prev_z+($lastz-$prev_z)*$partval*$i}]
		set reala [expr {$prev_a+($lasta-$prev_a)*$partval*$i}]
		set realb [expr {$prev_b+($lastb-$prev_b)*$partval*$i}]
		set realc [expr {$prev_c+($lastc-$prev_c)*$partval*$i}]

		if {![info exists a_matrix] || $reala != $prev_a} {
			set a_matrix [gcode_rotation_matrix $vector_ax1 $vector_ay1 $vector_az1 $reala]
		}
		if {$reala != 0.0} {
			set mat [list \
				[list [expr {$realx-$vector_ax0}]] \
				[list [expr {$realy-$vector_ay0}]] \
				[list [expr {$realz-$vector_az0}]] \
			]
			set mat [matrix_mult $a_matrix $mat]
			set realx [expr {[lindex [lindex $mat 0] 0]+$vector_ax0}]
			set realy [expr {[lindex [lindex $mat 1] 0]+$vector_ay0}]
			set realz [expr {[lindex [lindex $mat 2] 0]+$vector_az0}]
		}

		if {![info exists b_matrix] || $realb != $prev_b} {
			set b_matrix [gcode_rotation_matrix $vector_bx1 $vector_by1 $vector_bz1 $realb]
		}
		if {$realb != 0.0} {
			set mat [list \
				[list [expr {$realx-$vector_bx0}]] \
				[list [expr {$realy-$vector_by0}]] \
				[list [expr {$realz-$vector_bz0}]] \
			]
			set mat [matrix_mult $b_matrix $mat]
			set realx [expr {[lindex [lindex $mat 0] 0]+$vector_bx0}]
			set realy [expr {[lindex [lindex $mat 1] 0]+$vector_by0}]
			set realz [expr {[lindex [lindex $mat 2] 0]+$vector_bz0}]
		}

		if {![info exists c_matrix] || $realc != $prev_c} {
			set c_matrix [gcode_rotation_matrix $vector_cx1 $vector_cy1 $vector_cz1 $realc]
		}
		if {$realc != 0.0} {
			set mat [list \
				[list [expr {$realx-$vector_cx0}]] \
				[list [expr {$realy-$vector_cy0}]] \
				[list [expr {$realz-$vector_cz0}]] \
			]
			set mat [matrix_mult $c_matrix $mat]
			set realx [expr {[lindex [lindex $mat 0] 0]+$vector_cx0}]
			set realy [expr {[lindex [lindex $mat 1] 0]+$vector_cy0}]
			set realz [expr {[lindex [lindex $mat 2] 0]+$vector_cz0}]
		}

		lappend coordlist [list $realx $realy $realz $type $tool_width]
	}
	set prev_a $lasta
	set prev_b $lastb
	set prev_c $lastc
	set prev_x $lastx
	set prev_y $lasty
	set prev_z $lastz
}


proc gcode_parse_to_toolpath {channel {progresscb ""}} {
	global xoffset yoffset zoffset aoffset boffset coffset tool_width
	global prev_x prev_y prev_z
	global prev_a prev_b prev_c
	global lencomp lencomptool
	global diamcomp diamcomptool

	set pi 3.141592653589793236
	set max_arc_err 0.0001 ;# inches error

	set coordlist {}
	set coordsysnum 0
	for {set i 0} {$i <= 9} {incr i} {
		set coordsystems($i) [list 0.0 0.0 0.0 0.0 0.0 0.0]
	}
	set toolnum 1
	set gmode -1
	set speed 0.0
	set feedrate 1.0
	set feedmode "normal"
	set absmode 1
	set absarccenters 0
	set units "in"
	set plane "xy"
	set lencomptool 0
	set lencomp "none"
	set diamcomptool 0
	set diamcomp "none"
	set linenum 0
	set canned_feed 0.0
	set canned_return_level "orig"
	set tool_width 0.0

	set prev_x 0.0
	set prev_y 0.0
	set prev_z 0.0
	set prev_a 0.0
	set prev_b 0.0
	set prev_c 0.0

	set curra 0.0
	set currb 0.0
	set currc 0.0
	set currx 0.0
	set curry 0.0
	set currz 0.0

	set xnum $currx
	set ynum $curry
	set znum $currz
	set anum $curra
	set bnum $currb
	set cnum $currc
	set lnum 0
	set pnum 0.0
	set rnum 0.0

	set xoffset 0.0
	set yoffset 0.0
	set zoffset 0.0
	set aoffset 0.0
	set boffset 0.0
	set coffset 0.0

	gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
	while {![eof $channel]} {
		incr linenum
		if {[catch {gets $channel} line]} {
			break
		}
		if {$progresscb != ""} {
			set filepos [tell $channel]
			if {[catch {eval $progresscb $filepos}]} {
			    global errorInfo
				puts stderr $errorInfo
			}
		}
		set origline $line
		set line [regsub -all {[ 	]} $line ""]
		if {$line == ""} {
			continue
		}

		set have_x 0
		set have_y 0
		set have_z 0
		set have_a 0
		set have_b 0
		set have_c 0
		set have_l 0
		set have_p 0
		set have_r 0
		set inum 0.0
		set jnum 0.0
		set knum 0.0
		set do_set_origin 0
		set do_go_home 0
		set do_set_offsets 0
		set new_feed_rate $feedrate

		while {$line != ""} {
			set cmd [string toupper [string index $line 0]]
			set line [string trim [string range $line 1 end]]
			switch -exact -- $cmd {
				"#" {
					set parmnum [gcode_integer line $cmd $linenum]
					if {$parmnum < 1 || $parmnum > 5399} {
						error "Bad parameter number after # at line $linenum"
					}
					if {[string index $line 0] != "="} {
						error "Expected '=' after parameter number in line $linenum"
					}
					set line [string range $line 1 end]
					set val [gcode_number line $cmd $linenum]
					gcode_parameter_set $parmnum $val $linenum
				}
				"(" {
					if {![regexp -- {^([^)]*)\)(.*)$} $line dummy comment newstr]} {
						error "Unterminated comment in lin $linenum"
					}
					if {[regexp -nocase -- {\( ([0-9][0-9.]*) or [0-9/]* inches tool diam \)} $origline dummy diam]} {
						set tool_width $diam
					}
					set line $newstr
				}
				"A" {
					set anum [gcode_number line $cmd $linenum]
					set have_a 1
				}
				"B" {
					set bnum [gcode_number line $cmd $linenum]
					set have_b 1
				}
				"C" {
					set cnum [gcode_number line $cmd $linenum]
					set have_c 1
				}
				"D" {
					set diamcomptool [gcode_integer line $cmd $linenum]
				}
				"F" {
					set new_feed_rate [gcode_number line $cmd $linenum]
				}
				"G" {
					set gnum [gcode_number line $cmd $linenum]
					set gnum [format "%.1f" $gnum]
					switch -exact -- $gnum {
						0.0 - 1.0 - 2.0 - 3.0 {
							# movement command
							set gmode $gnum
						}
						4.0 {
							# Dwell.  Ignore this.
						}
						10.0 {
							set do_set_origin 1
						}
						17.0 {
							set plane "xy"
						}
						18.0 {
							set plane "xz"
						}
						19.0 {
							set plane "yz"
						}
						20.0 {
							set units "in"
						}
						21.0 {
							set units "mm"
						}
						28.0 {
							# return to home
							set do_go_home 1
						}
						30.0 {
							# return to secondary home
							set do_go_home 2
						}
						38.2 {
							# Straight-probe
							error "This parser cannot handle Straight-probe G38.2 codes."
						}
						40.0 {
							set diamcomp "none"
						}
						41.0 {
							set diamcomp "left"
						}
						42.0 {
							set diamcomp "right"
						}
						43.0 {
							set lencomp "plus"
						}
						49.0 {
							set lencomp "none"
						}
						53.0 {
							set coordsysnum 0
						}
						54.0 - 55.0 - 56.0 - 57.0 -
						58.0 - 59.0 {
							set coordsysnum [expr {int($gnum-53)}]
							gcode_parameter_set 5220 $coordsysnum $linenum
						}
						59.1 {
							set coordsysnum 7
							gcode_parameter_set 5220 $coordsysnum $linenum
						}
						59.2 {
							set coordsysnum 8
							gcode_parameter_set 5220 $coordsysnum $linenum
						}
						59.3 {
							set coordsysnum 9
							gcode_parameter_set 5220 $coordsysnum $linenum
						}
						61.0 - 61.1 - 64.0 {
							# Tool path control mode.  Ignore this.
						}
						80.0 {
							# Stop canned cycle mode
							set gmode -1
						}
						81.0 - 82.0 - 83.0 -
						84.0 - 85.0 - 86.0 -
						87.0 - 88.0 - 89.0 {
							# Canned movements
							set gmode $gnum
						}
						90.0 {
							set absmode 1
						}
						90.1 {
							set absarccenters 1
						}
						91.0 {
							set absmode 0
						}
						91.1 {
							set absarccenters 0
						}
						92.0 {
							set do_set_offsets 1
						}
						92.1 {
							set xoffset 0.0
							set yoffset 0.0
							set zoffset 0.0
							set aoffset 0.0
							set boffset 0.0
							set coffset 0.0
							gcode_parameter_set 5211 0.0 $linenum
							gcode_parameter_set 5212 0.0 $linenum
							gcode_parameter_set 5213 0.0 $linenum
							gcode_parameter_set 5214 0.0 $linenum
							gcode_parameter_set 5215 0.0 $linenum
							gcode_parameter_set 5216 0.0 $linenum
						}
						92.2 {
							set xoffset 0.0
							set yoffset 0.0
							set zoffset 0.0
							set aoffset 0.0
							set boffset 0.0
							set coffset 0.0
						}
						92.3 {
							set xoffset [gcode_parameter 5211 $linenum]
							set yoffset [gcode_parameter 5212 $linenum]
							set zoffset [gcode_parameter 5213 $linenum]
							set aoffset [gcode_parameter 5214 $linenum]
							set boffset [gcode_parameter 5215 $linenum]
							set coffset [gcode_parameter 5216 $linenum]
						}
						93.0 {
							set feedmode "inverse"
						}
						94.0 {
							set feedmode "normal"
						}
						98.0 {
							set canned_return_level "orig"
						}
						99.0 {
							set canned_return_level "R-point"
						}
						default {
							error "Unhandled code 'G$gnum' at line $linenum"
						}
					}
				}
				"H" {
					set lencomptool [gcode_integer line $cmd $linenum]
				}
				"I" {
					set inum [gcode_number line $cmd $linenum]
				}
				"J" {
					set jnum [gcode_number line $cmd $linenum]
				}
				"K" {
					set knum [gcode_number line $cmd $linenum]
				}
				"L" {
					set lnum [gcode_integer line $cmd $linenum]
					set have_l 1
				}
				"M" {
					set mcode [gcode_integer line $cmd $linenum]
					# Ignore M-codes.  Nothing there that afects path.
				}
				"N" {
					set linenum [gcode_integer line $cmd $linenum]
				}
				"P" {
					set pnum [gcode_number line $cmd $linenum]
					set have_p 1
				}
				"Q" {
					set canned_feed [gcode_number line $cmd $linenum]
				}
				"R" {
					set rnum [gcode_number line $cmd $linenum]
					set have_r 1
				}
				"S" {
					set speed [gcode_number line $cmd $linenum]
				}
				"T" {
					set toolnum [gcode_integer line $cmd $linenum]
				}
				"X" {
					set xnum [gcode_number line $cmd $linenum]
					set have_x 1
				}
				"Y" {
					set ynum [gcode_number line $cmd $linenum]
					set have_y 1
				}
				"Z" {
					set znum [gcode_number line $cmd $linenum]
					set have_z 1
				}
				default {
					error "Unhandled code '$cmd' at line $linenum"
				}
			}
		}
		if {$do_set_offsets} {
			set allgood 0
			if {$have_x} {
				set xoffset [expr {$xoffset+$currx-$xnum}]
				set currx $xnum
				gcode_parameter_set 5211 $xoffset $linenum
				set allgood 1
			}
			if {$have_y} {
				set yoffset [expr {$yoffset+$curry-$ynum}]
				set curry $ynum
				gcode_parameter_set 5212 $yoffset $linenum
				set allgood 1
			}
			if {$have_z} {
				set zoffset [expr {$zoffset+$currz-$znum}]
				set currz $znum
				gcode_parameter_set 5213 $zoffset $linenum
				set allgood 1
			}
			if {$have_a} {
				set aoffset [expr {$aoffset+$curra-$anum}]
				set curra $anum
				gcode_parameter_set 5214 $aoffset $linenum
				set allgood 1
			}
			if {$have_b} {
				set boffset [expr {$boffset+$currb-$bnum}]
				set currb $bnum
				gcode_parameter_set 5215 $boffset $linenum
				set allgood 1
			}
			if {$have_c} {
				set coffset [expr {$coffset+$currc-$cnum}]
				set currc $cnum
				gcode_parameter_set 5216 $coffset $linenum
				set allgood 1
			}
			if {!$allgood} {
				error "No axis words given for G92 offset code at line $linenum."
			}
		} elseif {$do_set_origin} {
			if {!$have_l || $lnum != 2.0} {
				error "Expected L2 after G10 at line $linenum"
			}
			if {!$have_p || $pnum < 1.0 || $pnum > 9.0 || $pnum != floor($pnum)} {
				error "Expected integer P value between 1 and 9 after G10 at line $linenum"
			}
			set parmbase [expr {5221+20*($pnum-1)}]
			if {$have_x} {
				gcode_parameter_set [expr {$parmbase+0}] $xnum $linenum
			}
			if {$have_y} {
				gcode_parameter_set [expr {$parmbase+1}] $ynum $linenum
			}
			if {$have_z} {
				gcode_parameter_set [expr {$parmbase+2}] $znum $linenum
			}
			if {$have_a} {
				gcode_parameter_set [expr {$parmbase+3}] $anum $linenum
			}
			if {$have_b} {
				gcode_parameter_set [expr {$parmbase+4}] $bnum $linenum
			}
			if {$have_c} {
				gcode_parameter_set [expr {$parmbase+5}] $cnum $linenum
			}
		} else {
			if {$do_go_home == 1} {
				gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "rapid" $linenum

				set xnum [gcode_parameter 5161 $linenum]
				set ynum [gcode_parameter 5162 $linenum]
				set znum [gcode_parameter 5163 $linenum]
				set anum [gcode_parameter 5164 $linenum]
				set bnum [gcode_parameter 5165 $linenum]
				set cnum [gcode_parameter 5166 $linenum]

				gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "rapid" $linenum

				set currx $xnum
				set curry $ynum
				set currz $znum
				set curra $anum
				set currb $bnum
				set currc $cnum
			} elseif {$do_go_home == 2} {
				gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "rapid" $linenum

				set xnum [gcode_parameter 5181 $linenum]
				set ynum [gcode_parameter 5182 $linenum]
				set znum [gcode_parameter 5183 $linenum]
				set anum [gcode_parameter 5184 $linenum]
				set bnum [gcode_parameter 5185 $linenum]
				set cnum [gcode_parameter 5186 $linenum]

				gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "rapid" $linenum

				set currx $xnum
				set curry $ynum
				set currz $znum
				set curra $anum
				set currb $bnum
				set currc $cnum
			} else {
				switch -exact -- $gmode {
					0.0 {
						# Rapid linear move
						if {!$absmode} {
							set xnum [expr {$currx+$xnum}]
							set ynum [expr {$curry+$ynum}]
							set znum [expr {$currz+$znum}]
							set anum [expr {$curra+$anum}]
							set bnum [expr {$currb+$bnum}]
							set cnum [expr {$currc+$cnum}]
						}
						gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "rapid" $linenum
						set currx $xnum
						set curry $ynum
						set currz $znum
						set curra $anum
						set currb $bnum
						set currc $cnum
					}
					1.0 {
						# Linear interpolation
						if {!$absmode} {
							set xnum [expr {$currx+$xnum}]
							set ynum [expr {$curry+$ynum}]
							set znum [expr {$currz+$znum}]
							set anum [expr {$curra+$anum}]
							set bnum [expr {$currb+$bnum}]
							set cnum [expr {$currc+$cnum}]
						}
						gcode_toolpath_append coordlist $xnum $ynum $znum $anum $bnum $cnum "linear" $linenum

						set currx $xnum
						set curry $ynum
						set currz $znum
						set curra $anum
						set currb $bnum
						set currc $cnum
					}
					2.0 - 3.0 {
						# Clockwise helical interpolation
						if {$gmode == "2.0"} {
							set htype "helicalcw"
						} else {
							set htype "helicalccw"
						}
						if {!$absmode} {
							set xnum [expr {$currx+$xnum}]
							set ynum [expr {$curry+$ynum}]
							set znum [expr {$currz+$znum}]
							set anum [expr {$curra+$anum}]
							set bnum [expr {$currb+$bnum}]
							set cnum [expr {$currc+$cnum}]
						}
						if {$plane == "xy"} {
							if {$have_r} {
								set radius $rnum
								set d [expr {hypot($ynum-$curry,$xnum-$currx)/2.0}]
								if {$d > $radius} {
									error "Radius of arc is too small for distance between points in line $linenum"
								}
								set ang [expr {atan2($ynum-$curry,$xnum-$currx)-($pi/2.0)}]
								set h [expr {sqrt($radius*$radius-$d*$d)}]
								set mx [expr {($xnum+$currx)/2.0}]
								set my [expr {($ynum+$curry)/2.0}]
								set cx [expr {$h*cos($ang)+$mx}]
								set cy [expr {$h*sin($ang)+$my}]
							} else {
								if {!$absarccenters} {
									set cx [expr {$currx+$inum}]
									set cy [expr {$curry+$jnum}]
									set radius [expr hypot($inum,$jnum)]
								} else {
									set cx $inum
									set cy $jnum
									set radius [expr hypot($cx-$currx,$cy-$curry)]
								}
							}
							set startang [expr {atan2($curry-$cy,$currx-$cx)}]
							set endang [expr {atan2($ynum-$cy,$xnum-$cx)}]
							if {$gmode == "2.0"} {
								if {$endang>=$startang} {
									set startang [expr {$startang+(2.0*$pi)}]
								}
							} else {
								if {$endang<=$startang} {
									set endang [expr {$endang+(2.0*$pi)}]
								}
							}
							if {$radius < 0.05} {
								set arc_res [expr {15.0*$pi/180.0}]
							} else {
								set arc_res [expr {2.0*acos(($radius-$max_arc_err)/$radius)}]
							}
							set steps [expr {1.0+int(abs($endang-$startang)/$arc_res)}]
							set stepang [expr {($endang-$startang)/$steps}]
							set stepz [expr {($znum-$currz)/$steps}]
							set stepa [expr {($anum-$curra)/$steps}]
							set stepb [expr {($bnum-$currb)/$steps}]
							set stepc [expr {($cnum-$currc)/$steps}]
							set arcz $currz
							set arca $curra
							set arcb $currb
							set arcc $currc
							set ang $startang
							while {1} {
								set arcx [expr {$cx+$radius*cos($ang)}]
								set arcy [expr {$cy+$radius*sin($ang)}]
								gcode_toolpath_append coordlist $arcx $arcy $arcz $anum $bnum $cnum $htype $linenum

								if {abs($ang-$endang) < 0.00001} {
									break
								}
								set arcz [expr {$arcz+$stepz}]
								set arca [expr {$arca+$stepa}]
								set arcb [expr {$arca+$stepb}]
								set arcc [expr {$arca+$stepc}]
								set ang [expr {$ang+$stepang}]
							}
						} elseif {$plane == "yz"} {
							if {$have_r} {
								set radius $rnum
								set d [expr {hypot($znum-$currz,$ynum-$curry)/2.0}]
								if {$d > $radius} {
									error "Radius of arc is too small for distance between points in line $linenum"
								}
								set ang [expr {atan2($znum-$currz,$ynum-$curry)-($pi/2.0)}]
								set h [expr {sqrt($radius*$radius-$d*$d)}]
								set my [expr {($ynum+$curry)/2.0}]
								set mz [expr {($znum+$currz)/2.0}]
								set cy [expr {$h*cos($ang)+$my}]
								set cz [expr {$h*sin($ang)+$mz}]
							} else {
								if {!$absarccenters} {
									set cy [expr {$curry+$jnum}]
									set cz [expr {$currz+$knum}]
									set radius [expr hypot($jnum,$knum)]
								} else {
									set cy $jnum
									set cz $knum
									set radius [expr hypot($cy-$curry,$cz-$currz)]
								}
							}
							set startang [expr {atan2($currz-$cz, $curry-$cy)}]
							set endang [expr {atan2($znum-$cz, $ynum-$cy)}]
							if {$gmode == "2.0"} {
								if {$endang>=$startang} {
									set startang [expr {$startang+(2.0*$pi)}]
								}
							} else {
								if {$endang<=$startang} {
									set endang [expr {$endang+(2.0*$pi)}]
								}
							}
							if {$radius < 0.05} {
								set arc_res [expr {15.0*$pi/180.0}]
							} else {
								set arc_res [expr {2.0*acos(($radius-$max_arc_err)/$radius)}]
							}
							set steps [expr {1.0+int(abs($endang-$startang)/$arc_res)}]
							set stepang [expr {($endang-$startang)/$steps}]
							set stepz [expr {($xnum-$currx)/$steps}]
							set stepa [expr {($anum-$curra)/$steps}]
							set stepb [expr {($bnum-$currb)/$steps}]
							set stepc [expr {($cnum-$currc)/$steps}]
							set arcx $currx
							set arca $curra
							set arcb $currb
							set arcc $currc
							set ang $startang
							while {1} {
								set arcy [expr {$cy+$radius*cos($ang)}]
								set arcz [expr {$cz+$radius*sin($ang)}]
								gcode_toolpath_append coordlist $arcx $arcy $arcz $anum $bnum $cnum $htype $linenum

								if {abs($ang-$endang) < 0.00001} {
									break
								}
								set arcx [expr {$arcx+$stepx}]
								set arca [expr {$arca+$stepa}]
								set arcb [expr {$arca+$stepb}]
								set arcc [expr {$arca+$stepc}]
								set ang [expr {$ang+$stepang}]
							}
						} else {
							# XZ plane
							if {$have_r} {
								set radius $rnum
								set d [expr {hypot($znum-$currz,$xnum-$currx)/2.0}]
								if {$d > $radius} {
									error "Radius of arc is too small for distance between points in line $linenum"
								}
								set ang [expr {atan2($znum-$currz,$xnum-$currx)-($pi/2.0)}]
								set h [expr {sqrt($radius*$radius-$d*$d)}]
								set mx [expr {($xnum+$currx)/2.0}]
								set mz [expr {($znum+$currz)/2.0}]
								set cx [expr {$h*cos($ang)+$mx}]
								set cz [expr {$h*sin($ang)+$mz}]
							} else {
								if {!$absarccenters} {
									set cx [expr {$currx+$inum}]
									set cz [expr {$currz+$knum}]
									set radius [expr hypot($inum,$knum)]
								} else {
									set cx $inum
									set cz $knum
									set radius [expr hypot($cx-$currx,$cz-$currz)]
								}
							}
							set startang [expr {atan2($currz-$cz, $currx-$cx)}]
							set endang [expr {atan2($znum-$cz, $xnum-$cx)}]
							if {$gmode == "2.0"} {
								if {$endang>=$startang} {
									set startang [expr {$startang+(2.0*$pi)}]
								}
							} else {
								if {$endang<=$startang} {
									set endang [expr {$endang+(2.0*$pi)}]
								}
							}
							if {$radius < 0.05} {
								set arc_res [expr {15.0*$pi/180.0}]
							} else {
								set arc_res [expr {2.0*acos(($radius-$max_arc_err)/$radius)}]
							}
							set steps [expr {1.0+int(abs($endang-$startang)/$arc_res)}]
							set stepang [expr {($endang-$startang)/$steps}]
							set stepy [expr {($ynum-$curry)/$steps}]
							set stepa [expr {($anum-$curra)/$steps}]
							set stepb [expr {($bnum-$currb)/$steps}]
							set stepc [expr {($cnum-$currc)/$steps}]
							set arcy $curry
							set arca $curra
							set arcb $currb
							set arcc $currc
							set ang $startang
							while {1} {
								set arcx [expr {$cx+$radius*cos($ang)}]
								set arcz [expr {$cz+$radius*sin($ang)}]
								gcode_toolpath_append coordlist $arcx $arcy $arcz $anum $bnum $cnum $htype $linenum

								if {abs($ang-$endang) < 0.00001} {
									break
								}
								set arcy [expr {$arcy+$stepy}]
								set arca [expr {$arca+$stepa}]
								set arcb [expr {$arca+$stepb}]
								set arcc [expr {$arca+$stepc}]
								set ang [expr {$ang+$stepang}]
							}
						}

						set currx $xnum
						set curry $ynum
						set currz $znum
						set curra $anum
						set currb $bnum
						set currc $cnum
					}
					81.0 - 82.0 - 83.0 -
					84.0 - 85.0 - 86.0 -
					87.0 - 88.0 - 89.0 {
						# Various drilling or boring cycles.
						for {set i 0} {$i < $lnum} {incr i} {
							if {$plane == "xy"} {
								if {$absmode} {
									set currx $xnum
									set curry $ynum
									set retractz $rnum
									set botz $znum
								} else {
									set currx [expr {$currx+$xnum}]
									set curry [expr {$curry+$ynum}]
									set retractz [expr {$currz+$rnum}]
									set botz [expr {$currz+$znum}]
								}
								if {$realrz > $realz} {
									gcode_toolpath_append coordlist $currx $curry $retractz $curra $currb $currc "rapid" $linenum
								}
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
								gcode_toolpath_append coordlist $currx $curry $botz $curra $currb $currc "drill" $linenum
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
							} elseif {$plane == "xz"} {
								if {$absmode} {
									set currx $xnum
									set currz $znum
									set retracty $rnum
									set boty $ynum
								} else {
									set currx [expr {$currx+$xnum}]
									set currz [expr {$currz+$znum}]
									set retracty [expr {$curry+$rnum}]
									set boty [expr {$curry+$ynum}]
								}
								if {$realrz > $realz} {
									gcode_toolpath_append coordlist $currx $retracty $currz $curra $currb $currc "rapid" $linenum
								}
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
								gcode_toolpath_append coordlist $currx $boty  $currz $curra $currb $currc "drill" $linenum
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
							} else {
								if {$absmode} {
									set curry $ynum
									set currz $znum
									set retractx $rnum
									set botx $xnum
								} else {
									set curry [expr {$curry+$ynum}]
									set currz [expr {$currz+$znum}]
									set retractx [expr {$currx+$rnum}]
									set botx [expr {$currx+$xnum}]
								}
								if {$realrz > $realz} {
									gcode_toolpath_append coordlist $retractx $curry $currz $curra $currb $currc "rapid" $linenum
								}
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
								gcode_toolpath_append coordlist $botx  $curry $currz $curra $currb $currc "drill" $linenum
								gcode_toolpath_append coordlist $currx $curry $currz $curra $currb $currc "rapid" $linenum
							}
						}
					}
				}
			}
		}
		gcode_parameters_commit
	}
	return $coordlist
}



#################################################################################################
# Plotting routines
#################################################################################################

proc gcode_3d_plot {base toolpathvar {ispreview 0}} {
	global xrot yrot zrot minx maxx miny maxy minz maxz scaleval scalepcnt rendertime
	set isocanv $base.isocanv
	set ysize [winfo reqwidth $isocanv]
	upvar $toolpathvar toolpath
	set scalmat [list \
		[list [expr {$scaleval*($scalepcnt/100.0)}] 0 0 0] \
		[list 0 [expr {-$scaleval*($scalepcnt/100.0)}] 0 0] \
		[list 0 0 [expr {$scaleval*($scalepcnt/100.0)}] 0] \
		[list 0 0 0 1] \
	]
	set mat $scalmat

	set xrotmat [list \
		[list 1 0 0 0] \
		[list 0 [expr {cos($xrot)}] [expr {-sin($xrot)}] 0] \
		[list 0 [expr {sin($xrot)}] [expr {cos($xrot)}] 0] \
		[list 0 0 0 1] \
	]
	set mat [matrix_mult $mat $xrotmat]

	set yrotmat [list \
		[list [expr {cos($yrot)}] 0 [expr {sin($yrot)}] 0] \
		[list 0 1 0 0] \
		[list [expr {-sin($yrot)}] 0 [expr {cos($yrot)}] 0] \
		[list 0 0 0 1] \
	]
	set mat [matrix_mult $mat $yrotmat]

	set zrotmat [list \
		[list [expr {cos($zrot)}] [expr {-sin($zrot)}] 0 0] \
		[list [expr {sin($zrot)}] [expr {cos($zrot)}] 0 0] \
		[list 0 0 1 0] \
		[list 0 0 0 1] \
	]
	set mat [matrix_mult $mat $zrotmat]

	set transmat [list \
		[list 1 0 0 [expr {-($maxx+$minx)/2.0}]] \
		[list 0 1 0 [expr {-($maxy+$miny)/2.0}]] \
		[list 0 0 1 [expr {-($maxz+$minz)/2.0}]] \
		[list 0 0 0 1] \
	]
	set mat [matrix_mult $mat $transmat]

	set a1 [lindex [lindex $mat 0] 0]
	set a2 [lindex [lindex $mat 0] 1]
	set a3 [lindex [lindex $mat 0] 2]
	set a4 [expr {[lindex [lindex $mat 0] 3]+$ysize/2.0}]
	set b1 [lindex [lindex $mat 1] 0]
	set b2 [lindex [lindex $mat 1] 1]
	set b3 [lindex [lindex $mat 1] 2]
	set b4 [expr {[lindex [lindex $mat 1] 3]+$ysize/2.0}]
	set isopath {}
	$isocanv delete all
	set backx [expr {$ysize-4}]
	set backy [expr {$ysize-4}]
	$isocanv create rectangle 0 0 $backx $backy -fill white -outline white
	if {$ispreview && $rendertime > 0.25} {
		set boxdata {}
		lappend boxdata $minx $miny $minz  $maxx $miny $minz  red
		lappend boxdata $maxx $miny $minz  $maxx $maxy $minz  black
		lappend boxdata $maxx $maxy $minz  $minx $maxy $minz  black
		lappend boxdata $minx $maxy $minz  $minx $miny $minz  green3

		lappend boxdata $minx $miny $minz  $minx $miny $maxz  blue
		lappend boxdata $maxx $miny $minz  $maxx $miny $maxz  black
		lappend boxdata $maxx $maxy $minz  $maxx $maxy $maxz  black
		lappend boxdata $minx $maxy $minz  $minx $maxy $maxz  black

		lappend boxdata $minx $miny $maxz  $maxx $miny $maxz  black
		lappend boxdata $maxx $miny $maxz  $maxx $maxy $maxz  black
		lappend boxdata $maxx $maxy $maxz  $minx $maxy $maxz  black
		lappend boxdata $minx $maxy $maxz  $minx $miny $maxz  black

		lappend boxdata $minx $miny $minz  $maxx $maxy $minz  grey
		lappend boxdata $maxx $miny $minz  $minx $maxy $minz  grey
		lappend boxdata $minx $miny $maxz  $maxx $maxy $maxz  cyan
		lappend boxdata $maxx $miny $maxz  $minx $maxy $maxz  cyan

		set linewidth 1.0
		foreach {x0 y0 z0 x1 y1 z1 color} $boxdata {
			set isox0 [expr {$a1*$x0+$a2*$y0+$a3*$z0+$a4}]
			set isoy0 [expr {$b1*$x0+$b2*$y0+$b3*$z0+$b4}]
			set isox1 [expr {$a1*$x1+$a2*$y1+$a3*$z1+$a4}]
			set isoy1 [expr {$b1*$x1+$b2*$y1+$b3*$z1+$b4}]
			$isocanv create line $isox0 $isoy0 $isox1 $isoy1 -fill $color -width $linewidth -dash "."
		}
	} else {
		set ox 0.0
		set oy 0.0
		set oz 0.0
		set otype "rapid"
		foreach pointdata $toolpath {
			foreach {x y z type tool_width} $pointdata break
			if {$type != $otype} {
				if {$otype == "rapid"} {
					set color cyan
					set dash "."
				} else {
					set color black
					set dash {}
				}
				if {[llength $isopath] >= 4} {
					$isocanv create line $isopath -fill $color -dash $dash
				}
				set isopath {}
				set isox [expr {$a1*$ox+$a2*$oy+$a3*$oz+$a4}]
				set isoy [expr {$b1*$ox+$b2*$oy+$b3*$oz+$b4}]
				lappend isopath $isox $isoy
			}
			set isox [expr {$a1*$x+$a2*$y+$a3*$z+$a4}]
			set isoy [expr {$b1*$x+$b2*$y+$b3*$z+$b4}]
			lappend isopath $isox $isoy
			set ox $x
			set oy $y
			set oz $z
			set otype $type
		}
		if {$otype == "rapid"} {
			set color cyan
			set dash "."
		} else {
			set color black
			set dash {}
		}
		if {[llength $isopath] >= 4} {
			$isocanv create line $isopath -fill $color -dash $dash
		}

		set midx [expr {($minx+$maxx)/2.0}]
		set midy [expr {($miny+$maxy)/2.0}]
		set midz [expr {($minz+$maxz)/2.0}]

		set axisdata {}
		#lappend axisdata $midx $midy $midz  $maxx $midy $midz  red
		#lappend axisdata $midx $midy $midz  $midx $maxy $midz  green3
		#lappend axisdata $midx $midy $midz  $midx $midy $maxz  blue
		lappend axisdata 0.0   0.0   0.0    1.0   0.0   0.0    red
		lappend axisdata 0.0   0.0   0.0    0.0   1.0   0.0    green3
		lappend axisdata 0.0   0.0   0.0    0.0   0.0   1.0    blue

		set linewidth 2.0
		foreach {x0 y0 z0 x1 y1 z1 color} $axisdata {
			set isox0 [expr {$a1*$x0+$a2*$y0+$a3*$z0+$a4}]
			set isoy0 [expr {$b1*$x0+$b2*$y0+$b3*$z0+$b4}]
			set isox1 [expr {$a1*$x1+$a2*$y1+$a3*$z1+$a4}]
			set isoy1 [expr {$b1*$x1+$b2*$y1+$b3*$z1+$b4}]
			if {hypot($isox1-$isox0,$isoy1-$isoy0) > 12.0} {
				$isocanv create line $isox0 $isoy0 $isox1 $isoy1 -fill $color -arrow last -width $linewidth -tags axis
			} else {
				$isocanv create line $isox0 $isoy0 $isox1 $isoy1 -fill $color -width $linewidth -tags axis
			}
		}
	}
	$isocanv raise all
	set bbox [$isocanv bbox all]
	if {$bbox != {}} {
		$isocanv configure -scrollregion $bbox
	}
}


proc gcode_3d_rot_start {isocanv x y} {
	global xrot_base yrot_base rot_dragging
	set xrot_base $x
	set yrot_base $y
	set rot_dragging 1
}


proc gcode_3d_rot_drag {base toolpathvar x y} {
	global xrot_base yrot_base xrot yrot zrot rot_dragging
	if {!$rot_dragging} {
		return
	}
	upvar $toolpathvar toolpath
	set isocanv $base.isocanv
	set ysize [winfo reqheight $isocanv]
	set dx [expr {$x-$xrot_base}]
	set dy [expr {$y-$yrot_base}]
	set dcx0 [expr {$xrot_base-$ysize/2.0}]
	set dcy0 [expr {$yrot_base-$ysize/2.0}]
	set dcx1 [expr {$x-$ysize/2.0}]
	set dcy1 [expr {$y-$ysize/2.0}]
	set ang0 [expr {atan2($dcy0,$dcx0)}]
	set ang1 [expr {atan2($dcy1,$dcx1)}]
	set dang [expr {$ang1-$ang0}]
	set dr [expr {hypot($dcy1,$dcx1)-hypot($dcy0,$dcx0)}]
	set xrot_base $x
	set yrot_base $y
	set pi 3.141592653589793236
	set xrot [expr {$xrot+0.5*$dy*$pi/180.0}]
	set yrot [expr {$yrot+0.5*$dx*$pi/180.0}]
	#set xrot [expr {$xrot+0.5*$dr*sin($ang0)*$pi/180.0}]
	#set yrot [expr {$yrot+0.5*$dr*cos($ang0)*$pi/180.0}]
	if {hypot($dcy0,$dcx0) > 100.0} {
		set zrot [expr {$zrot-$dang}]
	}
	gcode_3d_plot $base toolpath 1
}


proc gcode_3d_rot_end {base toolpathvar x y} {
	global rot_dragging
	upvar $toolpathvar toolpath
	set rot_dragging 0
	gcode_3d_plot $base toolpath
}


proc gcode_3d_scroll_horiz {base val} {
	if {$val < 0} {
		$base.isocanv xview scroll 1 units
	} else {
		$base.isocanv xview scroll -1 units
	}
}


proc gcode_3d_scroll_vert {base val} {
	if {$val < 0} {
		$base.isocanv yview scroll 1 units
	} else {
		$base.isocanv yview scroll -1 units
	}
}


proc gcode_3d_view_top {base} {
	gcode_3d_rotate_to $base 75 0.0 0.0 0.0
}


proc gcode_3d_view_front {base} {
	gcode_3d_rotate_to $base 75 -90.0 0.0 0.0
}


proc gcode_3d_view_side {base} {
	gcode_3d_rotate_to $base 75 -90.0 0.0 -90.0
}


proc gcode_3d_view_iso {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 -30.0
}


proc gcode_3d_view_iso2 {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 60.0
}


proc gcode_3d_view_iso3 {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 150.0
}


proc gcode_3d_view_iso4 {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 240.0
}


proc gcode_3d_view_iso5 {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 90.0
}


proc gcode_3d_view_iso6 {base} {
	gcode_3d_rotate_to $base 75 -45.0 0.0 0.0
}


proc gcode_3d_rotate_to {base scale_t xrot_t yrot_t zrot_t} {
	global xrot_targ yrot_targ zrot_targ scale_targ anim_pid
	set pi 3.141592653589793236
	set scale_targ $scale_t
	set xrot_targ [expr {$xrot_t*$pi/180.0}]
	set yrot_targ [expr {$yrot_t*$pi/180.0}]
	set zrot_targ [expr {$zrot_t*$pi/180.0}]
	if {[info exists anim_pid]} {
		after cancel $anim_pid
	}
	set anim_pid [after idle gcode_3d_rotate_anim $base]
}


proc gcode_3d_rotate_anim {base} {
	global scalepcnt scales scale_targ
	global xrot yrot zrot
	global xrot_targ yrot_targ zrot_targ
	global toolpath anim_pid rendertime

	set done 1
	if {$rendertime > 0.125} {
		set nextscale $scale_targ
		set scalepcnt $scale_targ
	} else {
		set nextscale [expr {($scalepcnt+$scale_targ)/2.0}]
	}
	if {$scale_targ < $nextscale} {
		foreach sc $scales {
			if {$sc < $nextscale} {
				set scalepcnt $sc
			}
		}
		set done 0
	} elseif {$scale_targ > $scalepcnt} {
		foreach sc $scales {
			if {$sc > $nextscale} {
				set scalepcnt $sc
				break
			}
		}
		set done 0
	}

	set dxrot [expr {$xrot_targ-$xrot}]
	set dyrot [expr {$yrot_targ-$yrot}]
	set dzrot [expr {$zrot_targ-$zrot}]
	if {$rendertime < 0.125} {
		if {abs($dxrot) > 0.01} {
			set dxrot [expr {$dxrot*0.667}]
			set done 0
		}
		if {abs($dyrot) > 0.01} {
			set dyrot [expr {$dyrot*0.667}]
			set done 0
		}
		if {abs($dzrot) > 0.01} {
			set dzrot [expr {$dzrot*0.667}]
			set done 0
		}
	}
	set xrot [expr {$xrot+$dxrot}]
	set yrot [expr {$yrot+$dyrot}]
	set zrot [expr {$zrot+$dzrot}]
	gcode_3d_plot $base toolpath
	update

	if {!$done} {
		set anim_pid [after idle gcode_3d_rotate_anim $base]
	} else {
		if {[info exists anim_pid]} {
			unset anim_pid
		}
	}
}


proc gcode_3d_scaling_change {base val x y} {
	global scalepcnt scales scaleval toolpath scalepid
	set isocanv $base.isocanv
	set scalenum [lsearch -exact $scales $scalepcnt]
	if {$val > 0} {
		if {$scalenum > 1} {
			incr scalenum -1
		}
	} else {
		if {$scalenum < [llength $scales]-1} {
			incr scalenum 1
		}
	}
	set scalepcnt [lindex $scales $scalenum]

	foreach {ix0 ix1} [$isocanv xview] break
	foreach {iy0 iy1} [$isocanv yview] break
	gcode_3d_plot $base toolpath 1
	foreach {ix2 ix3} [$isocanv xview] break
	foreach {iy2 iy3} [$isocanv yview] break

	set fullx [winfo reqwidth $isocanv]
	set fully [winfo reqwidth $isocanv]
	set px [expr {$x/double($fullx)}]
	set py [expr {$y/double($fullx)}]
	set dx0 [expr {$ix1-$ix0}]
	set dy0 [expr {$iy1-$iy0}]
	set dx1 [expr {$ix3-$ix2}]
	set dy1 [expr {$iy3-$iy2}]
	set vcx [expr {$ix0+$dx0*$px}]
	set vcy [expr {$iy0+$dy0*$py}]
	set nx [expr {$vcx-$dx1*$px}]
	set ny [expr {$vcy-$dy1*$py}]
	$isocanv xview moveto $nx
	$isocanv yview moveto $ny

	if {[info exists scalepid]} {
		after cancel $scalepid
	}
	set scalepid [after 500 gcode_3d_plot $base toolpath \; unset scalepid]
}


proc gcode_3view_plot {base toolpathvar} {
	set stime [clock clicks -milliseconds]

	global xrot yrot zrot
	global minx maxx miny maxy minz maxz
	global scalepcnt scaleval
	global scales rendertime
	global rot_dragging filename
	global xsize ysize zsize

	set xycanv $base.xycanv
	set xzcanv $base.xzcanv
	set yzcanv $base.yzcanv

	upvar $toolpathvar toolpath

	set zlevels 32

	set aspect [expr {($maxy-$miny)/($maxx-$minx)}]
	set canv_aspect [expr {[winfo reqheight $xycanv]/[winfo reqwidth $xycanv]}]
	if {$aspect < $canv_aspect} {
		set scaleval [expr {([winfo reqwidth $xycanv]-40.0)/($maxx-$minx)}]
	} else {
		set scaleval [expr {([winfo reqheight $xycanv]-40.0)/($maxy-$miny)}]
	}
	set dzmax [expr {abs($maxz-$minz)*$scaleval/(0.0+$zlevels)}]

	wm title $base "Backtrace of [file tail $filename]"

	# Plot 3 views
	set ox 0.0
	set oy 0.0
	set otype "rapid"

	set xoff1 [expr {$xsize/2.0-(($maxx+$minx)*$scaleval/2.0)}]
	set yoff1 [expr {$ysize/2.0-(($maxy+$miny)*$scaleval/2.0)}]

	set xoff2 [expr {$xsize/2.0-(($maxx+$minx)*$scaleval/2.0)}]
	set yoff2 [expr {$zsize/2.0-(($maxz+$minz)*$scaleval/2.0)}]

	set xoff3 [expr {$ysize/2.0-(($maxy+$miny)*$scaleval/2.0)}]
	set yoff3 [expr {$zsize/2.0-(($maxz+$minz)*$scaleval/2.0)}]

	$xycanv delete all
	$xzcanv delete all
	$yzcanv delete all

	set xypath {}
	set xzpath {}
	set yzpath {}
	set prevzq ""
	foreach pointdata $toolpath {
		foreach {x y z type tool_width} $pointdata break
		set zquant [expr {int((0.0+$zlevels)*($z-$minz)/($maxz-$minz))}]
		set x [expr {$x*$scaleval}]
		set y [expr {$y*$scaleval}]
		set z [expr {$z*$scaleval}]
		if {$prevzq == ""} {
			set prevzq $zquant
		}
		if {($type != $otype || $zquant!=$prevzq) && [llength $xypath]>2} {
			if {$otype == "rapid"} {
				set color cyan
				set color3 cyan
				set dash "."
			} else {
				set color black
				set color3 black
				set dash {}
				if {$tool_width > 0.0} {
					set cval [expr {191-int(191*abs(($oz-$minz)/($maxz-$minz)/$scaleval))}]
					set color2 [format "#%02x%02x%02x" $cval $cval $cval]
					set cval [expr {($cval+192-64)%192}]
					set color3 [format "#%02x%02x%02x" $cval $cval $cval]
					$xycanv create line $xypath -fill $color2 -width [expr {$scaleval*$tool_width}] -capstyle round -joinstyle round -tags "zq$prevzq"
				}
			}
			$xycanv create line $xypath -fill $color3 -dash $dash -tags path
			$xzcanv create line $xzpath -fill $color -dash $dash
			$yzcanv create line $yzpath -fill $color -dash $dash
			set prevzq $zquant
			set xypath {}
			set xzpath {}
			set yzpath {}
			lappend xypath [expr {$ox+$xoff1}] [expr {$ysize-($oy+$yoff1)}]
			lappend xzpath [expr {$ox+$xoff2}] [expr {$zsize-($oz+$yoff2)}]
			lappend yzpath [expr {$oy+$xoff3}] [expr {$zsize-($oz+$yoff3)}]
		}
		lappend xypath [expr {$x+$xoff1}] [expr {$ysize-($y+$yoff1)}]
		lappend xzpath [expr {$x+$xoff2}] [expr {$zsize-($z+$yoff2)}]
		lappend yzpath [expr {$y+$xoff3}] [expr {$zsize-($z+$yoff3)}]
		set ox $x
		set oy $y
		set oz $z
		set otype $type
	}
	if {$otype == "rapid"} {
		set color cyan
		set color3 cyan
		set dash "."
	} else {
		set color black
		set color3 black
		set dash {}
		if {$tool_width > 0.0} {
			set cval [expr {191-int(191*abs(($z-$minz)/($maxz-$minz)/$scaleval))}]
			set color2 [format "#%02x%02x%02x" $cval $cval $cval]
			set cval [expr {($cval+192-64)%192}]
			set color3 [format "#%02x%02x%02x" $cval $cval $cval]
			$xycanv create line $xypath -fill $color2 -width [expr {$scaleval*$tool_width}] -capstyle round -joinstyle round -tags "zq$zquant"
		}
	}
	$xycanv create line $xypath -fill $color3 -dash $dash -tags path
	$xzcanv create line $xzpath -fill $color -dash $dash
	$yzcanv create line $yzpath -fill $color -dash $dash

	$xycanv create line $xoff1 [expr {$ysize-$yoff1}] [expr {$scaleval+$xoff1}] [expr {$ysize-($yoff1)}] -arrow last -fill red -tags axis -width 2
	$xycanv create line $xoff1 [expr {$ysize-$yoff1}] $xoff1 [expr {$ysize-($scaleval+$yoff1)}] -arrow last -fill green -tags axis -width 2
	$xzcanv create line $xoff2 [expr {$zsize-$yoff2}] [expr {$scaleval+$xoff2}] [expr {$zsize-($yoff2)}] -arrow last -fill red -tags axis -width 2
	$xzcanv create line $xoff2 [expr {$zsize-$yoff2}] $xoff2 [expr {$zsize-($scaleval+$yoff2)}] -arrow last -fill blue -tags axis -width 2
	$yzcanv create line $xoff3 [expr {$zsize-$yoff3}] [expr {$scaleval+$xoff3}] [expr {$zsize-($yoff3)}] -arrow last -fill green -tags axis -width 2
	$yzcanv create line $xoff3 [expr {$zsize-$yoff3}] $xoff3 [expr {$zsize-($scaleval+$yoff3)}] -arrow last -fill blue -tags axis -width 2

	$xycanv raise all
	for {set zq $zlevels} {$zq >= 0} {incr zq -1} {
		$xycanv raise "zq$zq"
	}
	$xycanv raise path
	$xycanv raise axis
	$xzcanv raise axis
	$yzcanv raise axis
	$xzcanv raise all
	$yzcanv raise all

	puts stdout "plotting: [expr {([clock clicks -milliseconds]-$stime)/1000.0}] sec"
}


proc gcode_file_parse {file toolpathvar {progresscb ""}} {
	set stime [clock clicks -milliseconds]

	upvar $toolpathvar toolpath

	global filename
	global minx maxx miny maxy minz maxz

	set filename $file

	set f [open $file "r"]
	if {$progresscb != ""} {
		seek $f 0 end
		set endfilepos [tell $f]
		seek $f 0 start
		lappend progresscb $endfilepos
	}
	set toolpath [gcode_parse_to_toolpath $f $progresscb]
	catch {close $channel}

	puts stdout "parsing:  [expr {([clock clicks -milliseconds]-$stime)/1000.0}] sec"
	set stime [clock clicks -milliseconds]

	set minx ""
	set miny ""
	set minz ""
	set maxx ""
	set maxy ""
	set maxz ""
	foreach pointdata $toolpath {
		foreach {x y z type tool_width} $pointdata break
		if {$minx == "" || $minx > $x} {
			set minx $x
		}
		if {$miny == "" || $miny > $y} {
			set miny $y
		}
		if {$minz == "" || $minz > $z} {
			set minz $z
		}
		if {$maxx == "" || $maxx < $x} {
			set maxx $x
		}
		if {$maxy == "" || $maxy < $y} {
			set maxy $y
		}
		if {$maxz == "" || $maxz < $z} {
			set maxz $z
		}
	}
	if {$minx == ""} {
		error "No toolpath data found!"
	}
	if {abs($maxx-$minx) < 0.00001} {
		if {$maxy == $miny} {
			set maxx 1.0
		} else {
			set minx [expr {$minx-($maxy-$miny)/10.0}]
			set maxx [expr {$maxx+($maxy-$miny)/10.0}]
		}
	}
	if {abs($maxy-$miny) < 0.00001} {
		set miny [expr {$miny-($maxx-$minx)/10.0}]
		set maxy [expr {$maxy+($maxx-$minx)/10.0}]
	}

	puts stdout "scaling:  [expr {([clock clicks -milliseconds]-$stime)/1000.0}] sec"
}


proc gcode_3view_win_create {needsprefs} {
	global xrot yrot zrot
	global minx maxx miny maxy minz maxz
	global xsize ysize zsize
	global scalepcnt scaleval
	global toolpath scales rendertime
	global rot_dragging

	set scales [list 3.125 6.25 12.5 25 33 50 75 100 125 150 200 250 300 400 500 600 700 800 1000 1200 1400 1600]

	set base .viewwin
	set xsize 500
	set ysize 500
	set zsize 250

	toplevel $base -menu $base.menu
	wm protocol $base WM_DELETE_WINDOW "exit"

	set menubar [menu $base.menu -tearoff 0]
	set applemenu [menu $menubar.apple -tearoff 0]
	set filemenu  [menu $menubar.file -tearoff 0]
	$menubar add cascade -label "File" -menu $filemenu
	$filemenu add command -label "Open" -underline 0 -accelerator "Control+O" -command [list gcode_filemenu_open $base]

##### New code J.K.
	$filemenu add command -label "Toolfile" -underline 0 -accelerator "Control+T" -command [list gcode_toolfile_open $base]
	$filemenu add command -label "Print" -underline 0 -accelerator "Control+P" -command [list gcode_print_3view $base]
##### / J.K.

	if {$needsprefs} {
		$filemenu add command -label "Preferences..." -command [list gcode_prefs_dlog]
	}

	set xycanv  [canvas $base.xycanv  -width $xsize -height $ysize -highlightthickness 0 -borderwidth 1 -relief sunken -xscrollcommand [list $base.xyhscr set] -yscrollcommand [list $base.xyvscr set]]
	set xyhscr  [scrollbar $base.xyhscr -orient horizontal -command [list $xycanv xview]]
	set xyvscr  [scrollbar $base.xyvscr -orient vertical -command [list $xycanv yview]]

	set xzcanv  [canvas $base.xzcanv  -width $xsize -height $zsize -highlightthickness 0 -borderwidth 1 -relief sunken -xscrollcommand [list $base.xzhscr set] -yscrollcommand [list $base.xzvscr set]]
	set xzhscr  [scrollbar $base.xzhscr -orient horizontal -command [list $xzcanv xview]]
	set xzvscr  [scrollbar $base.xzvscr -orient vertical -command [list $xzcanv yview]]

	set yzcanv  [canvas $base.yzcanv  -width $ysize -height $zsize -highlightthickness 0 -borderwidth 1 -relief sunken -xscrollcommand [list $base.yzhscr set] -yscrollcommand [list $base.yzvscr set]]
	set yzhscr  [scrollbar $base.yzhscr -orient horizontal -command [list $yzcanv xview]]
	set yzvscr  [scrollbar $base.yzvscr -orient vertical -command [list $yzcanv yview]]

	set isocanv [canvas $base.isocanv -width $ysize -height $ysize -highlightthickness 0 -borderwidth 1 -relief sunken -xscrollcommand [list $base.isohscr set] -yscrollcommand [list $base.isovscr set]]
	set isohscr [scrollbar $base.isohscr -orient horizontal -command [list $isocanv xview]]
	set isovscr [scrollbar $base.isovscr -orient vertical -command [list $isocanv yview]]

	grid columnconfigure $base 0 -minsize 5
	grid columnconfigure $base 1 -weight 1
	grid columnconfigure $base 2 -minsize 5
	grid columnconfigure $base 3 -weight 1
	grid columnconfigure $base 4 -minsize 5
	grid rowconfigure $base 0 -minsize 5
	grid rowconfigure $base 1 -weight 1
	grid rowconfigure $base 2 -minsize 5
	grid rowconfigure $base 3 -weight 1
	grid rowconfigure $base 4 -minsize 5

	grid $xycanv  -row 1 -column 1 -sticky nsew
	grid $xyvscr  -row 1 -column 2 -sticky nsw
	grid $xyhscr  -row 2 -column 1 -sticky new

	grid $isocanv -row 1 -column 4 -sticky nsew
	grid $isovscr -row 1 -column 5 -sticky nsw
	grid $isohscr -row 2 -column 4 -sticky new

	grid $xzcanv  -row 4 -column 1 -sticky nsew
	grid $xzvscr  -row 4 -column 2 -sticky nsw
	grid $xzhscr  -row 5 -column 1 -sticky new

	grid $yzcanv  -row 4 -column 4 -sticky nsew
	grid $yzvscr  -row 4 -column 5 -sticky nsw
	grid $yzhscr  -row 5 -column 4 -sticky new

	set pi 3.141592653589793236
	set scalepcnt 75
	set xrot [expr {-45.0*$pi/180.0}]
	set yrot [expr   {0.0*$pi/180.0}]
	set zrot [expr {-30.0*$pi/180.0}]
	set mvx [expr {$ysize/2.0}]
	set rot_dragging 0

	bind $isocanv <ButtonPress-1>      [list gcode_3d_rot_start $base %x %y]
	bind $isocanv <Motion>             [list gcode_3d_rot_drag $base toolpath %x %y]
	bind $isocanv <ButtonRelease-1>    [list gcode_3d_rot_end $base toolpath %x %y]
	bind $isocanv <Command-Shift-MouseWheel> break
	bind $isocanv <Command-MouseWheel> [list gcode_3d_scaling_change $base %D %x %y]
	bind $isocanv <Shift-MouseWheel>   [list gcode_3d_scroll_horiz $base %D]
	bind $isocanv <MouseWheel>         [list gcode_3d_scroll_vert $base %D]
	bind $isocanv <Double-1>           [list gcode_3d_view_iso $base]

	bind $base <Key-1>  [list gcode_3d_view_top $base]
	bind $base <Key-2>  [list gcode_3d_view_front $base]
	bind $base <Key-3>  [list gcode_3d_view_side $base]
	bind $base <Key-4>  [list gcode_3d_view_iso $base]
	bind $base <Key-5>  [list gcode_3d_view_iso2 $base]
	bind $base <Key-6>  [list gcode_3d_view_iso3 $base]
	bind $base <Key-7>  [list gcode_3d_view_iso4 $base]
	bind $base <Key-8>  [list gcode_3d_view_iso5 $base]
	bind $base <Key-9>  [list gcode_3d_view_iso6 $base]

	bind $base <Key-Left>         [list gcode_3d_scroll_horiz $base  10]
	bind $base <Key-Right>        [list gcode_3d_scroll_horiz $base -10]
	bind $base <Key-Up>           [list gcode_3d_scroll_vert  $base  10]
	bind $base <Key-Down>         [list gcode_3d_scroll_vert  $base -10]
	bind $base <Command-Key-Up>   [list gcode_3d_scaling_change $base -1 $mvx $mvx]
	bind $base <Command-Key-Down> [list gcode_3d_scaling_change $base  1 $mvx $mvx]
	bind $base <Control-o>        [list $filemenu invoke "Open"]
##### New code J.K.
	bind $base <Control-t>        [list $filemenu invoke "Toolfile"]
	bind $base <Control-p>        [list $filemenu invoke "Print"]
##### / J.K.
	bind $base <Command-q>        exit
	bind $base <Command-w>        exit

	return $base
}


proc progress_create {win title caption} {
	toplevel $win
	wm resizable $win 0 0
	wm title $win $title
	label $win.caption -text $caption
	set canv [canvas $win.progbar -borderwidth 1 -relief solid -width 200 -height 20 -scrollregion {0 0 200 20}]
	$canv create rectangle 0 0 0 20 -width 0 -fill blue4 -tags progbar
	pack $win.caption -side top -anchor w
	pack $win.progbar -side top
	update idletasks
	raise $win
	return $win
}


proc progress_callback {win maxval currval} {
    set width [expr {int(round(200.0*$currval/$maxval))}]
	set oldwidth [expr {int(round([lindex [$win.progbar coords progbar] 2]))}]
	if {abs($width-$oldwidth) >= 2} {
		$win.progbar coords progbar 0 0 $width 20
		update idletasks
	}
}


proc progress_destroy {win} {
	if {[winfo exists $win]} {
		destroy $win
	}
}


proc gcode_load_and_show {file basewin} {
	global toolpath rendertime

	set toolpath {}

	set progwin [progress_create .progress "Backtracer progress" "Parsing..."]
	gcode_file_parse $file toolpath "progress_callback $progwin"
	progress_destroy $progwin

	gcode_3d_view_iso $basewin

	gcode_3view_plot $basewin toolpath

	set stime [clock clicks -milliseconds]
	gcode_3d_plot $basewin toolpath
	set rendertime [expr {([clock clicks -milliseconds]-$stime)/1000.0}]
	puts stdout "isoplot0: $rendertime sec"
	update
}


proc gcode_filemenu_open {base} {
    set filetypes {
        {{NC Files}	{.nc}	TEXT	}
	{{Any Files}	{*}	TEXT	}
    }
##### Modified code J.K.
	set file [tk_getOpenFile -defaultextension .nc -filetypes $filetypes \
		      -title "Open file" -parent $base]
##### / J.K.
	if {$file != ""} {
		gcode_load_and_show $file $base
	}
}


proc gcode_prefs_commit {base} {
	global vector_ax0 vector_ay0 vector_az0
	global vector_ax1 vector_ay1 vector_az1
	set pi 3.141592653589793236

	set afr $base.afr
	set xoff [$afr.xoff get]
	set yoff [$afr.yoff get]
	set zoff [$afr.zoff get]
	set xyang [$afr.xyang get]
	set tilt [$afr.tilt get]
	if {![string is double $xoff]} {
		tk_messageBox -type ok -icon error -message "X offset isn't a valid number." -parent $base
		return
	}
	if {![string is double $yoff]} {
		tk_messageBox -type ok -icon error -message "Y offset isn't a valid number." -parent $base
		return
	}
	if {![string is double $zoff]} {
		tk_messageBox -type ok -icon error -message "Z offset isn't a valid number." -parent $base
		return
	}
	if {![string is double $xyang]} {
		tk_messageBox -type ok -icon error -message "Orientation angle isn't a valid number." -parent $base
		return
	}
	if {![string is double $tilt]} {
		tk_messageBox -type ok -icon error -message "Tilt angle isn't a valid number." -parent $base
		return
	}

	set xyang [expr {$xyang*$pi/180.0}]
	set tilt [expr {$tilt*$pi/180.0}]

	set vector_ax0 $xoff
	set vector_ay0 $yoff
	set vector_az0 $zoff

	set vector_ax1 [expr {cos($xyang)*cos($tilt)}]
	set vector_ay1 [expr {sin($xyang)*cos($tilt)}]
	set vector_az1 [expr {sin($tilt)}]

	global filename
	if {[info exists filename]} {
		if {[file readable $filename]} {
			gcode_load_and_show $filename .viewwin
		}
	}
	destroy $base
}


proc gcode_prefs_dlog {} {
	set base ".prefs"

	global vector_ax0 vector_ay0 vector_az0
	global vector_ax1 vector_ay1 vector_az1
	set pi 3.141592653589793236

	toplevel $base
	wm title $base "Backtracer Preferences"

	set afr [labelframe $base.afr -text "A-Axis"]

	label $afr.xoffl -text "X offset"
	entry $afr.xoff  -width 20
	label $afr.yoffl -text "Y offset"
	entry $afr.yoff  -width 20
	label $afr.zoffl -text "Z offset"
	entry $afr.zoff  -width 20
	label $afr.xyangl -text "Orientation angle"
	entry $afr.xyang  -width 20
	label $afr.tiltl -text "Tilt angle"
	entry $afr.tilt  -width 20

	grid columnconfigure $afr 0 -minsize 10
	grid columnconfigure $afr 2 -minsize 5
	grid columnconfigure $afr 3 -weight 1
	grid columnconfigure $afr 4 -minsize 10
	grid rowconfigure $afr 0 -minsize 10
	grid rowconfigure $afr 2 -minsize 10
	grid rowconfigure $afr 4 -minsize 10
	grid rowconfigure $afr 6 -minsize 10
	grid rowconfigure $afr 8 -minsize 10
	grid rowconfigure $afr 10 -minsize 10 -weight 1
	grid x $afr.xoffl  x $afr.xoff  x -row 1 -sticky ew
	grid x $afr.yoffl  x $afr.yoff  x -row 3 -sticky ew
	grid x $afr.zoffl  x $afr.zoff  x -row 5 -sticky ew
	grid x $afr.xyangl x $afr.xyang x -row 7 -sticky ew
	grid x $afr.tiltl  x $afr.tilt  x -row 9 -sticky ew

	$afr.xoff insert end [format "%.4f" $vector_ax0]
	$afr.yoff insert end [format "%.4f" $vector_ay0]
	$afr.zoff insert end [format "%.4f" $vector_az0]
	set xyang [expr {atan2($vector_ay1,$vector_ax1)*180.0/$pi}]
	if {$xyang < 0.0} {
		set xyang [expr {$xyang+360.0}]
	}
	set xyang [format "%.2f" $xyang]
	$afr.xyang insert end $xyang
	set tilt [expr {asin($vector_az1)*180.0/$pi}]
	set tilt [format "%.2f" $tilt]
	$afr.tilt insert end $tilt

	button $base.cancel -text "Cancel" -width 6 -command "destroy $base"
	button $base.done -text "Done" -width 6 -command "gcode_prefs_commit $base"

	pack $afr -fill both -expand 1 -padx 10 -pady 10 -side top
	pack $base.cancel -padx 10 -pady 10 -side left
	pack $base.done -padx 10 -pady 10 -side left
	tkwait window $base
}




###############################################################################
# TESTING CODE FOLLOWS
###############################################################################
proc main {argc argv} {
	catch {set ::tk::mac::CGAntialiasLimit 0}
	global tcl_platform
	set needs_prefs 1
	catch {
		if {[tk windowingsystem] == "aqua"} {
			set needs_prefs 0
			namespace eval ::tk::mac {
				proc ShowPreferences {} {
					gcode_prefs_dlog
				}
			}
		}
	}
	gcode_tools_load "toolfile.txt"

	global vector_ax0 vector_ay0 vector_az0
	global vector_ax1 vector_ay1 vector_az1
	set vector_ax0 0.0
	set vector_ay0 0.0
	set vector_az0 0.0
	set vector_ax1 1.0
	set vector_ay1 0.0
	set vector_az1 0.0

	global vector_bx0 vector_by0 vector_bz0
	global vector_bx1 vector_by1 vector_bz1
	set vector_bx0 0.0
	set vector_by0 0.0
	set vector_bz0 0.0
	set vector_bx1 0.0
	set vector_by1 1.0
	set vector_bz1 0.0

	global vector_cx0 vector_cy0 vector_cz0
	global vector_cx1 vector_cy1 vector_cz1
	set vector_cx0 0.0
	set vector_cy0 0.0
	set vector_cz0 0.0
	set vector_cx1 0.0
	set vector_cy1 0.0
	set vector_cz1 1.0

	wm withdraw .
	set basewin [gcode_3view_win_create $needs_prefs]
	if {$argc == 1} {
		set file [lindex $argv 0]
		gcode_load_and_show $file $basewin
	}
}

main $argc $argv

# vim: set ts=4 sw=4 nowrap noexpandtab: settings
